//go:build (all || most || moderncsqlite) && !no_moderncsqlite

package internal

// Code generated by gen.go. DO NOT EDIT.

import (
	_ "github.com/xo/usql/drivers/moderncsqlite" // ModernC SQLite3 driver
)
