//go:build (all || most || maxcompute) && !no_maxcompute

package internal

// Code generated by gen.go. DO NOT EDIT.

import (
	_ "github.com/xo/usql/drivers/maxcompute" // Alibaba MaxCompute driver
)
