// Package internal provides a way to obtain information about which database
// drivers were included at build.
package internal

// Code generated by gen.go. DO NOT EDIT.

// KnownBuildTags returns a map of known driver names to its respective build
// tags.
func KnownBuildTags() map[string]string {
	return map[string]string{
		"adodb":         "adodb",         // github.com/mattn/go-adodb
		"athena":        "awsathena",     // github.com/uber/athenadriver/go
		"avatica":       "avatica",       // github.com/apache/calcite-avatica-go/v5
		"bigquery":      "bigquery",      // gorm.io/driver/bigquery/driver
		"cassandra":     "cql",           // github.com/MichaelS11/go-cql-driver
		"chai":          "chai",          // github.com/chaisql/chai/driver
		"clickhouse":    "clickhouse",    // github.com/ClickHouse/clickhouse-go/v2
		"cosmos":        "cosmos",        // github.com/btnguyen2k/gocosmos
		"couchbase":     "n1ql",          // github.com/couchbase/go_n1ql
		"csvq":          "csvq",          // github.com/mithrandie/csvq-driver
		"databend":      "databend",      // github.com/datafuselabs/databend-go
		"databricks":    "databricks",    // github.com/databricks/databricks-sql-go
		"duckdb":        "duckdb",        // github.com/marcboeker/go-duckdb
		"dynamodb":      "dynamodb",      // github.com/btnguyen2k/godynamo
		"exasol":        "exasol",        // github.com/exasol/exasol-driver-go
		"firebird":      "firebirdsql",   // github.com/nakagami/firebirdsql
		"flightsql":     "flightsql",     // github.com/apache/arrow/go/v17/arrow/flight/flightsql/driver
		"godror":        "godror",        // github.com/godror/godror
		"h2":            "h2",            // github.com/jmrobles/h2go
		"hive":          "hive",          // sqlflow.org/gohive
		"ignite":        "ignite",        // github.com/amsokol/ignite-go-client/sql
		"impala":        "impala",        // github.com/bippio/go-impala
		"maxcompute":    "maxcompute",    // sqlflow.org/gomaxcompute
		"moderncsqlite": "moderncsqlite", // modernc.org/sqlite
		"mymysql":       "mymysql",       // github.com/ziutek/mymysql/godrv
		"mysql":         "mysql",         // github.com/go-sql-driver/mysql
		"netezza":       "nzgo",          // github.com/IBM/nzgo/v12
		"odbc":          "odbc",          // github.com/alexbrainman/odbc
		"oracle":        "oracle",        // github.com/sijms/go-ora/v2
		"ots":           "ots",           // github.com/aliyun/aliyun-tablestore-go-sql-driver
		"pgx":           "pgx",           // github.com/jackc/pgx/v5/stdlib
		"postgres":      "postgres",      // github.com/lib/pq
		"presto":        "presto",        // github.com/prestodb/presto-go-client/presto
		"ql":            "ql",            // modernc.org/ql
		"ramsql":        "ramsql",        // github.com/proullon/ramsql/driver
		"sapase":        "tds",           // github.com/thda/tds
		"saphana":       "hdb",           // github.com/SAP/go-hdb/driver
		"snowflake":     "snowflake",     // github.com/snowflakedb/gosnowflake
		"spanner":       "spanner",       // github.com/googleapis/go-sql-spanner
		"sqlite3":       "sqlite3",       // github.com/mattn/go-sqlite3
		"sqlserver":     "sqlserver",     // github.com/microsoft/go-mssqldb
		"trino":         "trino",         // github.com/trinodb/trino-go-client/trino
		"vertica":       "vertica",       // github.com/vertica/vertica-sql-go
		"voltdb":        "voltdb",        // github.com/VoltDB/voltdb-client-go/voltdbclient
		"ydb":           "ydb",           // github.com/ydb-platform/ydb-go-sdk/v3
	}
}
