select 'testdata/sub/inc_test2.sql';

select 'from testdata/sub/inc_test2.sql, doing: \i inc_test_z.sql';

\i inc_test_z.sql

select 'from testdata/sub/inc_test2.sql, doing: \ir inc_test_z.sql';

\ir inc_test_z.sql
